/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.servicemix.cxfbc;

import java.io.ByteArrayOutputStream;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.InOptionalOut;
import javax.jbi.messaging.InOut;
import javax.jbi.messaging.MessageExchange;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.apache.cxf.binding.soap.SoapMessage;
import org.apache.cxf.binding.soap.interceptor.MustUnderstandInterceptor;
import org.apache.cxf.binding.soap.interceptor.ReadHeadersInterceptor;
import org.apache.cxf.endpoint.Endpoint;
import org.apache.cxf.interceptor.AttachmentInInterceptor;
import org.apache.cxf.interceptor.AttachmentOutInterceptor;
import org.apache.cxf.interceptor.Interceptor;
import org.apache.cxf.interceptor.StaxInInterceptor;
import org.apache.cxf.message.Attachment;
import org.apache.cxf.message.Exchange;
import org.apache.cxf.message.ExchangeImpl;
import org.apache.cxf.message.Message;
import org.apache.cxf.phase.PhaseChainCache;
import org.apache.cxf.phase.PhaseInterceptorChain;
import org.apache.cxf.phase.PhaseManager;
import org.apache.cxf.service.model.BindingOperationInfo;
import org.apache.cxf.service.model.EndpointInfo;
import org.apache.cxf.staxutils.DepthXMLStreamReader;
import org.apache.cxf.staxutils.StaxUtils;
import org.apache.cxf.transport.MessageObserver;
import org.apache.servicemix.JbiConstants;
import org.apache.servicemix.cxfbc.interceptors.JbiInWsdl1Interceptor;
import org.apache.servicemix.cxfbc.interceptors.JbiOperationInterceptor;
import org.apache.servicemix.jbi.jaxp.StringSource;

public class CxfBcProviderMessageObserver implements MessageObserver {
    ByteArrayOutputStream response = new ByteArrayOutputStream();

    boolean written;

    String contentType;

    private MessageExchange messageExchange;

    private CxfBcProvider providerEndpoint;

    public CxfBcProviderMessageObserver(MessageExchange exchange,
            CxfBcProvider providerEndpoint) {
        this.messageExchange = exchange;
        this.providerEndpoint = providerEndpoint;
    }

    public ByteArrayOutputStream getResponseStream() throws Exception {
        synchronized (this) {
            if (!written) {
                wait(1000000000);
            }
        }
        return response;
    }

    public String getResponseContentType() {
        return contentType;
    }

    public void onMessage(Message message) {
        try {
        	
            if (messageExchange.getStatus() != ExchangeStatus.ACTIVE) {
                return;
            }
            
            contentType = (String) message.get(Message.CONTENT_TYPE);
            SoapMessage soapMessage = new SoapMessage(message);
            
            

            // create XmlStreamReader
            EndpointInfo ei = providerEndpoint.getEndpointInfo();
            QName opeName = messageExchange.getOperation();
            
            
	        BindingOperationInfo boi = null;
            if (opeName == null) {
                // if interface only have one operation, may not specify the opeName in MessageExchange
                boi = ei.getBinding().getOperations().iterator().next();
            } else {
                boi = ei.getBinding().getOperation(opeName);   
            }
            
            if (boi.getOperationInfo().isOneWay()) {
                return;
            }
            
            if (contentType != null && contentType.toLowerCase().indexOf("multipart/related") == -1) {
            	System.out.println(" == The content type of response is not multipart ==" );
            	System.out.println(" == Creating XML Stream Reader == " );
            	XMLStreamReader xmlStreamReader = createXMLStreamReaderFromMessage(soapMessage);
            	soapMessage.setContent(XMLStreamReader.class, xmlStreamReader); 
            }else{
            	System.out.println(" == The content type of response is multipart not use XML Stream Reader ==" );
            	System.out.println(" == Creating XML Stream Reader == " );
            }
            
            soapMessage.put(org.apache.cxf.message.Message.REQUESTOR_ROLE, true);
            Exchange cxfExchange = new ExchangeImpl();
            soapMessage.setExchange(cxfExchange);

            cxfExchange.put(BindingOperationInfo.class, boi);
            cxfExchange.put(Endpoint.class, providerEndpoint.getCxfEndpoint());
            // create Interceptor chain
            PhaseChainCache inboundChainCache = new PhaseChainCache();
            PhaseManager pm = providerEndpoint.getBus().getExtension(
                    PhaseManager.class);
            
            
            List<Interceptor> inList = new ArrayList<Interceptor>();
            
           
            inList.add(new ReadHeadersInterceptor(providerEndpoint.getBus()));
            inList.add(new MustUnderstandInterceptor());
            inList.add(new AttachmentInInterceptor());
            inList.add(new StaxInInterceptor());
            inList.add(new ReadHeadersInterceptor(providerEndpoint.getBus()));
            inList.add(new JbiOperationInterceptor());
            inList.add(new JbiInWsdl1Interceptor());

            PhaseInterceptorChain inChain = inboundChainCache.get(pm
                    .getInPhases(), inList);
            inChain.add(providerEndpoint.getOutInterceptors());
            inChain.add(providerEndpoint.getOutFaultInterceptors());
            
            soapMessage.setInterceptorChain(inChain);
            inChain.doIntercept(soapMessage);
 
            
            if (boi.getOperationInfo().isOneWay()) {
                messageExchange.setStatus(ExchangeStatus.DONE);
            } else if (soapMessage.get("jbiFault") != null
                    && soapMessage.get("jbiFault").equals(true)) {
                Fault fault = messageExchange.createFault();
                fault.setContent(soapMessage.getContent(Source.class));
                messageExchange.setFault(fault);
            } else if (messageExchange instanceof InOut) {
            	
                NormalizedMessage msg = messageExchange.createMessage();
                msg.setContent(soapMessage.getContent(Source.class));
               
                Collection<Attachment> atts = soapMessage.getAttachments();
               
                if (atts != null){
                	for ( Attachment att : atts ){
                    	if (att != null){
                    		msg.addAttachment(att.getId(), att.getDataHandler());
                    	}else{
                    		System.out.println(" == Attachment is null == ");
                    	}
                    }
                }else{
                	System.out.println(" == No Attachments to add to Normalized Message == ");
                }
                
                messageExchange.setMessage(msg, "out");
            } else if (messageExchange instanceof InOptionalOut) {
                if (soapMessage.getContent(Source.class) != null) {
                    NormalizedMessage msg = messageExchange.createMessage();
                    msg.setContent(soapMessage.getContent(Source.class));
                    Collection<Attachment> atts = soapMessage.getAttachments();
                    if (atts != null){
                    	for ( Attachment att : atts ){
                        	if (att != null){
                        		msg.addAttachment(att.getId(), att.getDataHandler());
                        	}else{
                        		System.out.println(" == Attachment is null == ");
                        	}
                        }
                    }else{
                    	System.out.println(" == No Attachments to add to Normalized Message == ");
                    }
                    messageExchange.setMessage(msg, "out");
                } else {
                    messageExchange.setStatus(ExchangeStatus.DONE);
                }
            } else {
                messageExchange.setStatus(ExchangeStatus.DONE);

            }
            
            boolean txSync = messageExchange.getStatus() == ExchangeStatus.ACTIVE
                    && messageExchange.isTransacted()
                    && Boolean.TRUE.equals(messageExchange
                            .getProperty(JbiConstants.SEND_SYNC));
            
            if (txSync) {
            	
                providerEndpoint.getContext().getDeliveryChannel().sendSync(
                        messageExchange);
            } else {
            	
                providerEndpoint.getContext().getDeliveryChannel().send(
                        messageExchange);
            }

        } catch (Exception e) {
            e.printStackTrace();
        } finally {
            synchronized (this) {
                written = true;
                notifyAll();
            }
        }
    }

    public void onError(Exception ex) {
		try {
			if (messageExchange.getStatus() != ExchangeStatus.ACTIVE) {
				return;
			}

			if ((messageExchange instanceof InOut)
					|| (messageExchange instanceof InOptionalOut)) {
				NormalizedMessage msg = messageExchange.createMessage();

				// --------- Temporary patch ----------------
				StringBuffer response = new StringBuffer();
				response.append("<ERROR>");
				response.append("<![CDATA[");
				response.append(ex.getMessage());
				response.append("]]>");
				response.append("</ERROR>");

				msg.setContent(new StringSource(response.toString()));
				messageExchange.setMessage(msg, "out");
				// --------- End Temporary patch ----------------
			} else {
				messageExchange.setStatus(ExchangeStatus.DONE);
			}
			boolean txSync = messageExchange.getStatus() == ExchangeStatus.ACTIVE
					&& messageExchange.isTransacted()
					&& Boolean.TRUE.equals(messageExchange
							.getProperty(JbiConstants.SEND_SYNC));
			if (txSync) {
				providerEndpoint.getContext().getDeliveryChannel().sendSync(
						messageExchange);
			} else {
				providerEndpoint.getContext().getDeliveryChannel().send(
						messageExchange);
			}

		} catch (Exception e) {
			e.printStackTrace();
		} finally {
			synchronized (this) {
				written = true;
				notifyAll();
			}
		}
	}
    
    private XMLStreamReader createXMLStreamReaderFromMessage(Message message) {
        StreamSource bodySource = new StreamSource(message
                .getContent(InputStream.class));
        XMLStreamReader xmlReader = StaxUtils.createXMLStreamReader(bodySource);
        return xmlReader;
    }
    
    /** 3.2.2 CODE 
    private XMLStreamReader createXMLStreamReaderFromMessage(Message message) {
        XMLStreamReader xmlReader = null;
        StreamSource bodySource = new StreamSource(message
                .getContent(InputStream.class));
        xmlReader = StaxUtils.createXMLStreamReader(bodySource);
        
        findBody(message, xmlReader);
        
        return xmlReader;
    }
    
    private void findBody(Message message, XMLStreamReader xmlReader) {
        DepthXMLStreamReader reader = new DepthXMLStreamReader(xmlReader);
        try {
            int depth = reader.getDepth();
            int event = reader.getEventType();
            while (reader.getDepth() >= depth && reader.hasNext()) {
                QName name = null;
                if (event == XMLStreamReader.START_ELEMENT) {
                    name = reader.getName();
                }
                if (event == XMLStreamReader.START_ELEMENT && name.equals(((SoapMessage)message).getVersion().getBody())) {
                    reader.nextTag();
                    return;
                }
                event = reader.next();
            }
            return;
        } catch (XMLStreamException e) {
            throw new RuntimeException("Couldn't parse stream.", e);
        }
    }
    */
}